/*
 * An implementation of Rabin-Karp algorithm
 * See http://www.giaithuatlaptrinh.com/?p=290 for details
 *
 *  Created on: Aug 5, 2015
 *      Author: hunglv
 */

#include<time.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#define MAXN 100
#define MAXM 20
#define PRIME 1
#define COMPOSITE 0

typedef unsigned long long int LLU; // using unsigned long long type to avoid overflow

char T[MAXN];
char P[MAXM];



LLU mod_power(LLU a, LLU b, LLU p);
LLU random_prime(LLU N);
int miller_rabin_testing(int acc, LLU N);
LLU random64();
LLU gcd(LLU a, LLU b);
LLU mod_mul(LLU a, LLU b, LLU p);

int RabinKarpMatcher(int n,int m);


void main (void){
	char text[] = "dlcimglkblhcch";
	char pattern[] = "mglkblhc";
	int i = 0;
	T[0] = " ";
	for(i = 0; i < strlen(text); i++){
		T[i+1] = text[i];
	}
	P[0] = " ";
	for(i = 0; i < strlen(pattern); i++){
		P[i+1] = pattern[i];
	}
	int n = 14, m = 8;
	printf("%d\n", RabinKarpMatcher(n,m));
}


int RabinKarpMatcher(int n, int m){
	LLU base = (LLU) 26; // the alphabet is {a,b,c,d,...,z}
	LLU q = random_prime((LLU)(1<<30) - 1);// generating a random prime number, assume that m^2 log m <= 2^(30)
	LLU p = 0;
	LLU Ts[n-m+2];
	LLU sig = mod_power(base,m-1,q); // base^(m-1) \mod q
	Ts[1] = 0;
	int i = 0;
	for(i = 1; i <=m; i++){
		p = (P[i]-'a' + ((base*p)%q))%q;
		Ts[1] = (T[i]-'a'+ ((base*Ts[1])%q))%q;
	}
	for(i = 1; i <= n-m+1; i++){
		if(Ts[i] == p){
			if(isequal(i,m)){ // compare T[i,i+1,.., i+m-1] with P[1,2,...,m]
				return i;
			}
		} else {
			Ts[i+1] =  ((base*((Ts[i]  - ((sig*(T[i]-'a')) %q) + q) %q))%q + T[i+m]-'a')%q;
		}
	}
  return 0;
}

int isequal(int s, int m){
	int i = 1;
	while(T[s+i-1] == P[i] && i <= m){
		i++;
	}
	if(i > m) return 1;
	else return 0;
}

/*
 * generating a random 32 bit prime number using Miller-Rabin primality testing
 * See //www.giaithuatlaptrinh.com/?p=278 for details
 */
LLU random_prime(LLU N){
	LLU q = random64()%(N-2)+2;
	while(miller_rabin_testing(10, q) != PRIME){
		q = random64()%(N-2)+2;
	}
	return q;
}

/*
 * Miller-Rabin Primality testing algorithm
 * See //www.giaithuatlaptrinh.com/?p=278 for details
 */

int miller_rabin_testing(int acc, LLU N){
	LLU a = 0;
	int i = 0;
	for(i = 0; i <= acc; i++){
		a = random64()%(N-2)+ 2;
		if(gcd(a,N) != 1){
			return COMPOSITE;
		} else if (mod_power(a,N-1,N) != 1){
			return COMPOSITE;
		}else {
			if (witness(a,N) == COMPOSITE){
				return COMPOSITE;
			}
		}
	}
	return PRIME;
}
/*
 * gcd(LLU a, LLU b) return the greteast common divisor of two number a and b.
 */
LLU gcd(LLU a, LLU b){
	LLU tmp;
	while  (b != 0){
		tmp  = b;
		b = a % b;
		a = tmp;
	}
	return a;
}

/*
 * mod_power(LLU a, LLU b, LLU p) return a^b \mod q
 */
LLU mod_power(LLU a, LLU b, LLU p){
	if (b == 1){
		return a %p;
	} else {
		LLU x = mod_power(a, b/2, p);
		if((b& 1) == 0){ // b is even
			return mod_mul(x,x,p); // return  (x*x)%p;
		} else {
			return (mod_mul(mod_mul(x,x,p), a, p));// return (x*x*a)%p;
		}
	}
}
/*
 * this function compute (a*b) \mod p using (\log_2a + 1) space
 */

LLU mod_mul(LLU a, LLU b, LLU p){
	if (b == 1){
		return a %p;
	} else {
		LLU x = mod_mul(a, b/2, p);
		if((b& 1) == 0){ // b is even
			return (x+x)%p;
		} else {
			return (((x+x)%p)+a)%p;
		}
	}
}


int witness(LLU a,  LLU N){
	LLU m;
	int k;
	decompose(N-1,&k,&m);
	LLU B[k+1];
	B[0] = mod_power(a,m,N);
	if(B[0] == 1){
		return PRIME;
	} else {
		int i = 1;
		while(i <= k){
			B[i] =  mod_mul(B[i-1],B[i-1],N);// (B[i-1]*B[i-1])%N;
			if(B[i] == 1){
				if(B[i-1] == N-1){
					return PRIME;
				} else {
					return COMPOSITE;
				}
			}
			i++;
		}
	}
	return COMPOSITE;
}

void decompose(LLU p, int *k, LLU *m){
	int i = 0;
	while ((p& 1) == 0){ // p is even
		i++;
		p = p >> 1; // p = p /2
	}
	*k = i;
	*m = p;
}
/*
 * generating a random 64 bit number
 */
LLU random64(){
	LLU n  = 0;
	int l = rand()%(RAND_MAX-2)+2; // last 32 random bit
	int r = rand()%(RAND_MAX) + rand()%2; // first 32 random bit
	n = n | ((LLU)r);
	n = n << 32;
	n = n | ((LLU)l);
	return n;
}

